# Dashboard Visualization APIs

Dashboard Visualization is an React component with following props defined

```js
import { Component } from 'react';
import T from 'prop-types';

class BaseVisualization extends Component {
    static propTypes = {
        /**
         * width in pixels or string, defaults to 100%
         */
        width: T.oneOfType([T.string, T.number]),
        /**
         * height in pixels or string
         */
        height: T.oneOfType([T.string, T.number]),
        /**
         * If set to "true" there will be an indication that the visualization data is loading
         */
        loading: T.bool,
        /**
         * visualization formatting options
         */
        options: T.object,
        /**
         * A callback to update formatting options
         */
        onOptionsChange: T.func,
        /**
         * visual encoding configuration
         * to bind datasource fields to visual properties
         */
        encoding: T.object,
        /**
         * datasource state which include data and request params, object key indicate the datasource type.
         */
        dataSources: T.objectOf(
            T.shape({
                /**
                 * current request params
                 */
                requestParams: T.object,
                /**
                 * current dataset
                 */
                data: T.shape({
                    fields: T.array,
                    columns: T.array,
                }),
                /**
                 * meta data that came with the dataset
                 */
                meta: T.object,
            })
        ),
        /**
         * A callback to trigger event
         */
        onEventTrigger: T.func,
        /**
         * A callback to request new data with updated request params
         */
        onRequestParamsChange: T.func,
        /**
         * Custom style on visualization
         */
        style: T.object,
    };
    static defaultProps = {
        width: '100%',
        options: {},
        style: {},
        onOptionsChange: () => {},
        onEventTrigger: () => {},
        onRequestParamsChange: () => {},
    };
    /**
     * default data contract
     */
    static dataContract = {
        requiredDataSources: ['primary'],
        initialRequestParams: {
            primary: { offset: 0, count: 10000 },
        },
    };
    /**
     * focus can be implemented by the visualization to put focus on
     * certain elements of it when called
     */
    focus = () => {};
    render() {
        return null;
    }
}
```

## Basic Principle

### A visualization should

-   Honestly present data with specific formatting options.
-   Communicate to consumer with [Event](#/VisualizationEvent) via `onEventTrigger` callback.
-   Be predictable. Always render identical result given a fixed dataset and options.

### A visualization should NOT

-   Depends on globals.
-   Issue requests to fetch data.

## Visualization Data

Visualization reads the data from a dataSources object in json_cols format. For example:

```js
dataSources={{
    primary: {
        requestParams: { offset: 0, count: 20 },
        data: {
            fields: [{ name: 'component', groupby_rank: '0' }, { name: 'count' }],
            columns: [
                ['ApplicationLicense', 'DatabaseDirectoryManager', 'IndexWriter'],
                ['166238', '8904', '2327'],
            ],
        },
        meta: { totalCount: 100 },
    },
}}
```

### RequestParams

RequestParams define how does a visualization wants the data.

The following example indicate that this is the first 20 results of a given DataSource

```js
{
    requestParams: { offset: 0, count: 20 },
    data: {
        fields: [{ name: 'component', groupby_rank: '0' }, { name: 'count' }],
        columns: [
            ['ApplicationLicense', 'DatabaseDirectoryManager', 'IndexWriter'],
            ['166238', '8904', '2327'],
        ],
    },
    meta: { totalCount: 100 },
}
```

Visualizations invoke the `onRequestParamsChange` callback to request data with updated RequestParams (if necessary)

```js
this.props.onRequestParamsChange('primary', {
    ...this.props.dataSources.primary.requestParams,
    offset: 20,
    count: 20,
});
```

Checkout out a complete example _Update RequestParams_ from Table Visualization.

All available RequestParams are documented in DataSource module.

### DataSources

DataSource is a module that do the data fetching, see @splunk/datasources for more details.

-   While Visualization and DataSource are designed to work together, Visualization itself can be used as a standalone React component.
-   Each Visualization can be powered by multiple DataSources - 0 to 1 `primary` DataSource and 0 to n secondary DataSource

### Data Contract

Visualizations expose a static property called `dataContract` to indicate which datasources and visual encodings are supported. The following snippet is an example data contract:

```js
static dataContract = {
    requiredDataSources: ['primary'],
    optionalDataSources: ['annotation'],
    encoding: {
        x: {
            isRequired: true,
            type: ['time', 'number', 'string'],
            default: 'primary[0]',
        },
        y: {
            isRequired: true,
            type: ['number', 'string'],
            default: 'primary[1]',
        },
    },
    initialRequestParams: {
        offset: 0,
        count: 10000,
    }
}
```

From the data contract above UDF is able to infer that the visualization requires one datasource called `primary`, and supports an optional datasource called `annotation`. `encoding` specifies that the visualization can handle an x, and y data field, where it expects certain data types.
Last, it exposes initialRequestParams to let datasources know what parameters to use for the initial fetching of data.

#### Supported DataSources

The data contract contains a property `requiredDataSources`, and `optionalDataSources` to indicate what kind of DataSources it supports. For example, Table only supports primary DataSource:

```js
static dataContract = {
    requiredDataSources: ['primary'],
    //...
}
```

while Line Chart supports an additional annotation DataSource

```js
static dataContract = {
    requiredDataSources: ['primary'],
    optionalDataSources: ['annotation']
    //...
}
```

#### Encodings

Visualizations can expose fields that they understand via an `encoding` property in the data contract. Each field can either be required or optional and can have certain type limitations.

By exposing the expected fields, data to visualization field-bindings can be defined explicitly and data can be passed down to the visualization in a format that it knows how to handle.

See @splunk/visualization-encoding-parsers

#### Initial RequestParams

In order to retrieve the first batch of data, the initial RequestParams needs to be figured out before the Visualization gets rendered. The `initialRequestParams` property in the `dataContract` serves this purpose.

initialRequestParams can be an object

```js
initialRequestParams: {
    offset: 0,
    count: 10000,
}
```

or a function that takes the visualization options and returns the initial RequestParams

```js
initialRequestParams: (options = {}) => ({
    offset: 0,
    count: options.count || 20,
});
```

### Meta

Meta object tells visualization extra information besides the data.

For example:

```js
{
    totalCount: 100, // total results
    progress: 50, // 0-100, current search progress
}
```

It's up to the viz implementation to decide whether to use them. Checkout Meta from DataSource package for details.

## Size

If either `width` or `height` is provided, the visualization will render within the specified dimensions and handle the size change.

## Placeholder

You can define a default placeholder for visualizations using the `withPlaceholder` component. In this library, most visualizations have default placeholders with a few exceptions such as Text viz. For custom visualizations, a customized placeholder can be provided to `withPlaceholder`. For example:

```js
import React, { Component } from 'react';
import CustomPlaceholder from '...';
import { withPlaceholder } from '@splunk/dashboard-visualizations/utils/enhancer';

class CustomViz extends Component {
    ...
}

export default withPlaceholder({
    placeholder: <CustomPlaceholder />
})(CustomViz);
```

## Locale, timezone, and i18n

These should be defined independently for each visualization. You can use `@splunk/moment` and `@splunk/ui-utils/i18n` to define them.
